indexing
	description: "Mathematical objects that use higher-order functionals"

deferred class
	MML_FUNCTIONALS [G]

inherit
	ANY
		export
			{NONE}
				all
		undefine
			copy, is_equal
		end

feature -- Function Iteration

	iterated (other : FUNCTION[ANY, TUPLE[G], G]; n : INTEGER) : FUNCTION[ANY, TUPLE[G], G] is
			-- The `n'-th iterate of `other'.
		require
			other_not_void : other /= Void
			need_positive_n : n > 0
		do
			Result := agent iterated_adapter (?, other, n)
		ensure
			result_not_void : Result /= Void
		end

feature -- Function Currying

	curried (other : FUNCTION[ANY, TUPLE[G, ANY], ANY]) : FUNCTION[ANY, TUPLE[G], FUNCTION[ANY, TUPLE[ANY], ANY]] is
			-- The curried form of `other'.
		require
			other_not_void : other /= Void
		do
			Result := agent curried_adapter (?, other)
		ensure
			result_not_void : Result /= Void
		end

feature -- Function Composition

	composed (first : FUNCTION[ANY, TUPLE[ANY], ANY]; second : FUNCTION[ANY, TUPLE[ANY], ANY]) : FUNCTION[ANY, TUPLE[ANY], ANY] is
			-- The composition of `first' and `second'.
		require
			first_not_void : first /= Void
			second_not_void : second /= Void
		do
			Result := agent composed_adapter (?, first,  second)
		ensure
			result_not_void : Result /= Void
		end

feature -- Boolean Operators

	negated (other : G; predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- The negation of `predicate'.
		require
			predicate_not_void : predicate /= Void
			other_not_void : other /= Void
		do
			Result := not predicate.item ([other])
		ensure
			definition_of_negated: Result.is_equal (not predicate.item ([other]))
		end

	ored (other : G; left: FUNCTION [ANY, TUPLE [G], BOOLEAN]; right: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- The disjunction of `left' and `right'.
		require
			left_not_void : left /= Void
			right_not_void : right /= Void
			other_not_void : other /= Void
		do
			Result := left.item ([other]) or right.item ([other])
		ensure
			definition_of_ored: Result.is_equal (left.item ([other]) or right.item ([other]))
		end

	xored (other : G; left: FUNCTION [ANY, TUPLE [G], BOOLEAN]; right: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- The exclusive disjunction of `left' and `right'.
		require
			left_not_void : left /= Void
			right_not_void : right /= Void
			other_not_void : other /= Void
		do
			Result := left.item ([other]) xor right.item ([other])
		ensure
			definition_of_xored: Result.is_equal (left.item ([other]) xor right.item ([other]))
		end

	anded (other : G; left: FUNCTION [ANY, TUPLE [G], BOOLEAN]; right: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- The conjunction of `left' and `right'.
		require
			left_not_void : left /= Void
			right_not_void : right /= Void
			other_not_void : other /= Void
		do
			Result := left.item ([other]) and right.item ([other])
		ensure
			definition_of_anded: Result.is_equal (left.item ([other]) and right.item ([other]))
		end

	implied (other : G; left: FUNCTION [ANY, TUPLE [G], BOOLEAN]; right: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- The implication of `left' and `right'.
		require
			left_not_void : left /= Void
			right_not_void : right /= Void
			other_not_void : other /= Void
		do
			Result := left.item ([other]) implies right.item ([other])
		ensure
			definition_of_implied: Result.is_equal (left.item ([other]) implies right.item ([other]))
		end

feature {NONE} -- Implementation

	composed_adapter (argument : ANY; first : FUNCTION[ANY, TUPLE[ANY], ANY]; second : FUNCTION[ANY, TUPLE[ANY], ANY]) : ANY is
			-- The adapter feature of `composed'.
		require
			first_not_void : first /= Void
			second_not_void : second /= Void
		do
			Result := first.item ([second.item ([argument])])
		ensure
			result_not_void : Result /= Void
		end

	curried_adapter (argument : G; other : FUNCTION[ANY, TUPLE[G, ANY], ANY]) : FUNCTION[ANY, TUPLE[ANY], ANY] is
			-- The adapter feature of `curried'.
		require
			other_not_void : other /= Void
		do
			Result := agent curried_result_adapter (?, argument, other)
		ensure
			result_not_void : Result /= Void
		end

	curried_result_adapter (second : ANY; first : G; other : FUNCTION[ANY, TUPLE[G, ANY], ANY]) : ANY is
			-- The result adapter feature of `curried'.
		require
			other_not_void : other /= Void
		do
			Result := other.item ([first, second])
		ensure
			result_not_void : Result /= Void
		end

	iterated_adapter (argument : G; other : FUNCTION[ANY, TUPLE[G], G]; n : INTEGER) : G is
			-- The adapter feature of `iterated'.
		require
			argument_not_void : argument /= Void
			other_not_void : other /= Void
			need_positive_n : n > 0
		do
			if n > 0 then
				Result := iterated_adapter (other.item ([argument]), other, n - 1)
			else
				Result := other.item([argument])
			end
		ensure
			result_not_void : Result /= Void
		end

end -- class MML_FUNCTIONALS