indexing
	description: "Default implementation of MML_SET[G]"

class
	MML_DEFAULT_SET [G]

inherit
	MML_SET[G]
		redefine
			copy
		end

create
	make_empty, make_from_element, make_from_set, make_from_other

feature {NONE} -- Initialization

	make_from_element (other : G) is
			-- Create a new set containing `other'.
		do
			make_empty
			extend (other)
		ensure then
			implementation_of_make_from_element : is_equal ((create {like Current}.make_empty).extended (other))
		end

	make_from_other (other : MML_SET[G]) is
			-- Create a new object containing the elements from `other'.
		do
			make_from_set (other)
		end

	make_from_set (other : MML_SET[G]) is
			-- Create a new set with elements from `other'.
		do
			make_empty
			other.do_all (agent extend (?))
		end

	make_empty is
			-- Create a new empty set.
		do
			create {ARRAYED_SET[G]} representation.make (0)
			representation.compare_objects
		ensure then
			representation_not_void: representation /= Void
			representation_compares_objects: representation.object_comparison
		end

feature -- Access

	any_element : G is
			-- An arbitrary element of `current'.
		local
			dummy: like representation
		do
			dummy := clone (representation)
			dummy.start

			Result := dummy.item
		ensure then
			result_not_void : Result /= Void
		end

	any_item (predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : G is
			-- An arbitrary element of `current' which satisfies `predicate'?
		do
			Result := filtered (predicate).any_element
		ensure then
			result_not_void : Result /= Void
		end

feature -- Conversion

	identity_pair : MML_PAIR[like Current, like Current] is
			-- The identity pair of `current'.
		do
			Result := create {MML_DEFAULT_PAIR[like Current, like Current]}.make_from (Current, Current)
		ensure then
			implementation_of_identity_pair : Result.is_equal (create {MML_DEFAULT_PAIR[like Current, like Current]}.make_from (Current, Current))
			result_not_void : Result /= Void
		end

	identity_pair_element (other : G) : MML_PAIR[G, G] is
			-- The element `other' as identity pair
		do
			create {MML_DEFAULT_PAIR[G, G]} Result.make_from (other, other)
		ensure then
			implementation_of_identity_pair : Result.is_equal (create {MML_DEFAULT_PAIR[G, G]}.make_from (other, other))
		end

	identity_relation : MML_RELATION[G, G] is
			-- The identity relation of `current'.
		local
			accumulator : MML_DEFAULT_RELATION[G, G]
		do
			create accumulator.make_empty
			mapped_to_identity (agent identity_pair_element (?)).do_all (agent accumulator.extend (?))

			Result := accumulator
		ensure then
			result_not_void : Result /= Void
		end

	power_set : MML_POWERSET[G] is
			-- The set `current' as power set.
		do
			create {MML_DEFAULT_POWERSET[G]} Result.make_from_element (Current)
		ensure then
			implementation_of_power_set : Result.is_equal (create {MML_DEFAULT_POWERSET[G]}.make_from_element (Current))
			result_not_void : Result /= Void
		end

	random_sequence : MML_SEQUENCE[G] is
			-- The set `current' as random sequence.
		local
			accumulator : MML_DEFAULT_SEQUENCE[G]
		do
			create accumulator.make_empty
			do_all (agent accumulator.append_element (?))

			Result := accumulator
		ensure then
			result_not_void : Result /= Void
		end

feature -- Status Report

	cardinality : INTEGER is
			-- The cardinality of `current'
		do
			Result := representation.count
		ensure then
			implementation_of_cardinality: Result.is_equal (representation.count)
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			Precursor (other)
			representation := clone (other.representation)
		end

feature -- Properties

	is_member (other: G) : BOOLEAN is
			-- Is `other' a member of `current'?
		do
			Result := representation.has (other)
		ensure then
			implementation_of_is_member: Result.is_equal (representation.has (other))
		end

feature -- Basic Operations

	cartesian_product (other : MML_SET[ANY]) : MML_SET[MML_PAIR[G, ANY]] is
			-- The cartesian product of `other' and `current'.
		local
			accumulator : MML_DEFAULT_SET[MML_PAIR[G, ANY]]
		do
			create accumulator.make_empty
			other.do_all (agent accumulate_cartesian_product (?, Current, accumulator))

			Result := accumulator
		ensure then
			result_not_void : Result /= Void
		end

	intersected (other: MML_SET[G]) : like Current is
			-- The intersection of `current' and `other'.
		do
			Result := clone (Current)
			do_if (agent Result.prune (?), agent negated (?, agent other.is_member (?)))
		ensure then
			result_not_void : Result /= Void
		end

	united (other: MML_SET[G]) : MML_SET[G] is
			-- The union of `current' and `other'.
		do
			Result := clone (Current)
			other.do_all (agent Result.extend (?))
		ensure then
			result_not_void : Result /= Void
		end

	subtracted (other: MML_SET[G]) : like Current is
			-- The difference of `current' and `other'.
		do
			Result := clone (Current)
			do_if (agent Result.prune (?), agent other.is_member (?))
		ensure then
			result_not_void : Result /= Void
		end

	extended (other: G) : like Current is
			-- The set `current' extended with `other'.
		do
			Result := clone (Current)
			Result.extend (other)
		ensure then
			result_not_void : Result /= Void
		end

	pruned (other: G) : like Current is
			-- The set `current' with `other' removed.
		do
			Result := clone (Current)
			Result.prune (other)
		ensure then
			result_not_void : Result /= Void
		end

feature -- Processing

	applied (action: PROCEDURE [ANY, TUPLE [G]]) : like Current is
			-- The object with `action' applied to every element of `current'.
		do
			Result := clone (Current)
			Result.do_all (action)
		ensure then
			result_not_void : Result /= Void
		end

feature -- Filtering

	filtered (predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : like Current is
			-- The object containing all elements in `current' not filtered by `predicate'.
		do
			create Result.make_empty
			do_if (agent Result.extend (?), predicate)
		ensure then
			result_not_void : Result /= Void
		end

feature -- Partition

	is_partition (other : MML_SET[MML_SET[G]]) : BOOLEAN is
			-- Is `other' a partition of `current'?
		do
			Result := (create {MML_DEFAULT_POWERSET[G]}.make_from_set (other)).is_generalized_partition (Current)
		ensure then
			implementation_of_partition : Result.is_equal ((create {MML_DEFAULT_POWERSET[G]}.make_from_set (other)).is_generalized_partition (Current))
			result_not_void : Result /= Void
		end

feature -- Quantifiers

	there_exists (predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- Does `current' contain an element which satisfies `predicate'?
		do
			Result := representation.there_exists (predicate)
		ensure then
			implementation_of_there_exists: Result.is_equal (representation.there_exists (predicate))
		end

	for_all (predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : BOOLEAN is
			-- Does `current' contain only elements which satisfy `predicate'?
		do
			Result := representation.for_all (predicate)
		ensure then
			implementation_of_for_all: Result.is_equal (representation.for_all (predicate))
		end

feature -- Mapping

	mapped_to_identity (mapping: FUNCTION [ANY, TUPLE [G], MML_PAIR[G, G]]) : MML_SET[MML_PAIR[G, G]] is
			-- The object containing all elements in `current' mapped using the `mapping' function.
		do
			create {MML_DEFAULT_SET[MML_PAIR[G, G]]} Result.make_empty
			do_all (agent accumulate_identity_pair (?, mapping, Result))
		ensure then
			result_not_void : Result /= Void
		end

feature {MML_ANY} -- Implementation

	do_all (action: PROCEDURE [ANY, TUPLE [G]]) is
			-- Apply `action' to every element of `current'.
			-- CAUTION: This feature is not pure
		do
			representation.do_all (action)
		end

	do_if (action: PROCEDURE [ANY, TUPLE [G]]; predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) is
			-- Apply `action' to every element of `current' that satisfies `predicate'.
			-- CAUTION: This feature is not pure
		do
			representation.do_if (action, predicate)
		end

	extend (other: G) is
			-- Ensure that representation includes `other'.
			-- CAUTION: This feature is not pure
		do
			representation.extend (other)
		end

	prune (other: G) is
			-- Ensure that representation does not include `other'.
			-- CAUTION: This feature is not pure
		do
			representation.prune (other)
		end

feature {MML_DEFAULT_SET} -- Implementation

	representation: ARRAYED_LIST[G]
		-- The internal representation

feature {NONE} -- Implementation

	accumulate_identity_pair (element : G; transformator : FUNCTION [ANY, TUPLE [G], MML_PAIR[G, G]]; accumulator : MML_EXTENDIBLE[MML_PAIR[G, G]]) is
			-- Accumulate the identity pair of `element' computed by `transformator' in `accumulator'.
			-- CAUTION: This feature is not pure
		require
			element_not_void : element /= Void
			transformator_not_void : transformator /= Void
			accumulator_not_void : accumulator /= Void
		do
			accumulator.extend (transformator.item ([element]))
		end

	accumulate_cartesian_product (element : ANY; processable : MML_PROCESSABLE [G]; accumulator : MML_EXTENDIBLE[MML_PAIR[G, ANY]]) is
			-- Accumulate the cartesian produce of `element' and `processable' in `accumulator'.
			-- CAUTION: This feature is not pure
		require
			element_not_void : element /= Void
			processable_not_void : processable /= Void
			accumulator_not_void : accumulator /= Void
		do
			processable.do_all (agent accumulate_cartesian_pair (?, element, accumulator))
		end

	accumulate_cartesian_pair (source : G; target : ANY; accumulator : MML_EXTENDIBLE[MML_PAIR[G, ANY]]) is
			-- Accumulate the cartesian pair of `source' and `target' in `accumulator'.
			-- CAUTION: This feature is not pure
		require
			source_not_void : source /= Void
			target_not_void : target /= Void
			accumulator_not_void : accumulator /= Void
		do
			accumulator.extend (create {MML_DEFAULT_PAIR[G, ANY]}.make_from (source, target))
		end

invariant
	representation_not_void: representation /= Void
	representation_compares_objects: representation.object_comparison

end -- class MML_DEFAULT_SET