indexing
	description: "Mathematical objects that contain elements of the same type in unspecified order"

deferred class
	MML_BAG[G]

inherit
	MML_FUNCTION[G, INTEGER]
		rename
			inversed as functional_inversed,
			relational_inversed as inversed,
			evaluated as multiplicity,
			is_defined as is_member,
			image as occurrences,
			has_mapping as has_multiplicity,
			intersected as set_intersected,
			united as set_united,
			difference as set_difference,
			subtracted as set_subtracted,
			relational_extended as set_extended,
			pruned as set_pruned,
			cardinality as set_cardinality,
			do_if as functional_do_if,
			do_all as functional_do_all,
			applied as functional_applied,
			for_all as functional_for_all,
			there_exists as functional_there_exists,
			negated as functional_negated,
			ored as functional_ored,
			xored as functional_xored,
			anded as functional_anded,
			implied as functional_implied,
			curried_adapter as functional_curried_adapter,
			curried_result_adapter as functional_curried_result_adapter,
			curried_implementation as functional_curried_implementation,
			iterated as functional_iterated,
			iterated_adapter as functional_iterated_adapter,
			untyped as functional_untyped,
			random_sequence as set_random_sequence
		export
			{MML_ANY}
				set_intersected, set_united, set_subtracted, set_extended, set_pruned, set_difference,
				is_subset, is_proper_subset, is_superset, is_proper_superset, functional_untyped,
				relational_intersected, relational_united, functional_inversed, set_random_sequence
		undefine
			equals
		select
			functional_for_all, functional_there_exists,
			functional_do_if, functional_do_all, functional_applied,
			functional_negated, functional_ored, functional_xored,
			functional_anded, functional_implied, functional_curried_adapter,
			functional_curried_implementation, functional_curried_result_adapter,
			functional_iterated, functional_iterated_adapter, functional_untyped
		end

	MML_TYPEABLE[MML_BAG[G]]

	MML_PROCESSABLE[G]

	MML_QUANTIFIABLE[G]
		rename
			composed as composed_implementation,
			curried as bag_curried_implementation,
			curried_adapter as bag_curried_adapter,
			curried_result_adapter as bag_curried_result_adapter,
			iterated_adapter as sequence_iterated_adapter,
			iterated as sequence_iterated
		undefine
			composed_adapter, composed_implementation
		end

feature {NONE} -- Initialization

	make_from_other (other : MML_BAG[G]) is
			-- Create a new object containing the elements from `other'.
		deferred
		end

feature -- Comparison

	equals (other: MML_BAG[G]): BOOLEAN is
			-- Is `other' mathematically equivalent to `current' ?
		deferred
		end

feature -- Type

	untyped : MML_BAG[ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

feature -- Conversion

	random_sequence : MML_SEQUENCE[G] is
			-- The bag `current' as random sequence.
		deferred
		ensure
			definition_of_random_sequence : Result.for_all (agent is_member (?)) and domain.is_equal (Result.range)
			same_cardinality : Result.count.is_equal (cardinality)
		end

feature -- Measurement

	cardinality : INTEGER is
			-- The cardinality of `current'
		deferred
		ensure
			non_negative_cardinality : Result >= 0
		end

feature -- Properties

	is_subbag (other: MML_BAG[G]) : BOOLEAN is
			-- Is `other' a subbag of `current'?
		require
			other_not_void : other /= Void
		do
			Result := domain.is_subset (other.domain) and then other.domain.for_all (agent is_included (?, other))
		ensure
			definition_of_subbag : Result.is_equal (domain.is_subset (other.domain) and other.domain.for_all (agent is_included (?, other)))
			valid_cardinality : Result.is_equal (cardinality >= other.cardinality)
		end

	is_superbag (other: MML_BAG[G]) : BOOLEAN is
			-- Is `other' a superbag of `current'?
		require
			other_not_void : other /= Void
		do
			Result := other.is_subbag (Current)
		ensure
			definition_of_superbag : Result.is_equal (other.is_subbag (Current))
			valid_cardinality : Result.is_equal (other.cardinality >= cardinality)
		end

	is_proper_superbag (other: MML_BAG[G]) : BOOLEAN is
			-- Is `other' a proper superbag of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_superbag (other) and not is_equal (other)
		ensure
			definition_of_proper_superbag : Result.is_equal (is_superbag (other) and not is_equal (other))
			valid_cardinality : Result.is_equal (other.cardinality > cardinality)
		end

	is_proper_subbag (other: MML_BAG[G]) : BOOLEAN is
			-- Is `other' a proper subbag of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_subbag (other) and not is_equal (other)
		ensure
			definition_of_proper_subbag : Result.is_equal (is_subbag (other) and not is_equal (other))
			valid_cardinality : Result.is_equal (other.cardinality < cardinality)
		end

feature -- Basic Operations

	intersected (other: MML_BAG[G]) : like Current is
			-- The intersection of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_intersection : Result.domain.is_equal (domain.intersected (other.domain))
		end

	united (other: MML_BAG[G]) : like Current is
			-- The union of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_union : Result.domain.is_equal (domain.united (other.domain))
		end

	subtracted, removed (other: MML_BAG[G]) : like Current is
			-- The difference of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			disjoint_subtracted_result : Result.intersected (other).is_empty
			valid_cardinality : Result.cardinality <= cardinality
		end

	difference (other: MML_BAG[G]) : like Current is
			-- The symmetric difference of `current' and `other'.
		require
			other_not_void : other /= Void
		do
			Result := united (other).subtracted (intersected (other))
		ensure
			definition_of_difference : Result.is_equal (united (other).subtracted (intersected (other)))
		end

	added (other: like Current) : like Current is
			-- The bag `current' with all elements from `other'.
		require
			other_not_void: other /= Void
		deferred
		ensure
			definition_of_extended: Result.domain.is_equal (domain.united (other.domain))
		end

	extended (other: G) : like Current is
			-- The bag `current' extended with an occurrence `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_extended : Result.is_member (other)
			one_element_more : is_member (other) implies (Result.multiplicity (other).is_equal (multiplicity (other) + 1))
			other_members_remain : domain.for_all (agent Result.is_member (?))
			other_multiplicities_remain : Result.domain_element_anti_restricted (other).range.is_equal (domain_element_anti_restricted (other).range)
		end

	pruned (other: G) : like Current is
			-- The bag `current' with an occurrence `other' removed.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_pruned : is_equal (Result.extended (other))
			one_element_less : is_member (other) and Result.is_member (other) implies Result.multiplicity (other).is_equal (multiplicity (other) - 1)
			other_members_remain : Result.domain.for_all (agent is_member (?))
			other_multiplicities_remain : Result.domain_element_anti_restricted (other).range.is_equal (domain_element_anti_restricted (other).range)
		end

feature {NONE} -- Implementation

	is_positive (number : INTEGER) : BOOLEAN is
			-- Is `number' positive?
		do
			Result := number > 0
		ensure
			definition_of_positive : Result.is_equal (number > 0)
		end

	is_included (other : G; bag : MML_BAG[G]) : BOOLEAN is
			-- Are all occurrences of `other' in `bag' also included in `current'?
		require
			bag_not_void : bag /= Void
			other_not_void : other /= Void
			other_member_bag : bag.is_member (other)
			other_member_current : is_member (other)
		do
			Result := bag.multiplicity (other) <= multiplicity (other)
		ensure
			definition_of_included : Result.is_equal (bag.multiplicity (other) <= multiplicity (other))
		end

invariant
	definition_of_bag : range.for_all (agent is_positive (?)) and domain.cardinality.is_equal (set_cardinality)
	definition_of_union : is_equal (united (Current))
	definition_of_intersection : is_equal (intersected (Current))

	non_negative_cardinality : cardinality >= 0
	definition_of_included : for_all (agent is_included (?, Current))

end -- class MML_BAG