indexing
	description: "Mathematical relations that relate each element of the domain set to at least one element in the range set"

deferred class
	MML_FUNCTION[G, H]

inherit
	MML_RELATION[G, H]
		rename
			intersected as relational_intersected,
			united as relational_united,
			is_member as has_mapping,
			inversed as relational_inversed,
			extended as relational_extended,
			curried as curried_implementation
		export
			{MML_ANY}
				relational_extended
		redefine
			untyped
		end

feature {NONE} -- Initialization

	make_from_other (other : MML_FUNCTION[G, H]) is
			-- Create a new object containing the elements from `other'.
		deferred
		end

	make_from_relation (other : MML_RELATION[G, H]) is
			-- Create a new function representing `other'.
		require
			other_not_void : other /= Void
			must_be_functional : other.domain.for_all (agent is_functional (?))
		deferred
		ensure
			definition_of_make_from_relation: other.is_equal (Current)
			same_members: other.for_all (agent has_mapping (?))
			same_cardinality: cardinality.is_equal (other.cardinality)
		end

feature -- Type

	untyped : MML_FUNCTION[ANY, ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

feature -- Status Report

	is_continuous : BOOLEAN is
			-- Is `current' continuous?
		do
			Result := not is_discrete
		ensure
			definition_of_continuous : Result.is_equal (not is_discrete)
		end

	is_discrete : BOOLEAN is
			-- Is `current' discrete?
		do
			Result := true
		end

	is_invertible : BOOLEAN is
			-- Is `current' invertible?
		do
			Result := is_injective
		ensure
			definition_of_invertible : is_injective implies Result
		end

	is_finite : BOOLEAN is
			-- Is `current' finite?
		do
			Result := true
		end

	is_partial : BOOLEAN is
			-- Is `current' partial?
		do
			Result := true
		end

	is_total : BOOLEAN is
			-- Is `current' total?
		do
			Result := not is_partial
		ensure
			definition_of_total : Result.is_equal (not is_partial)
		end

	is_injective : BOOLEAN is
			-- Is `current' injective?
		deferred
		end

	is_surjective : BOOLEAN is
			-- Is `current' surjective?
		do
			Result := true
		end

	is_bijective : BOOLEAN is
			-- Is `current' bijective?
		do
			Result := is_total and then is_injective and then is_surjective
		ensure
			definition_of_bijective : Result.is_equal (is_injective and is_surjective and is_total)
		end

	is_defined (argument : G) : BOOLEAN is
			-- Is `current' defined for `argument'?
		require
			argument_not_void : argument /= Void
		deferred
		ensure
			defined_on_empty_set : is_empty implies not Result
		end

feature -- Basic Operations

	intersected (other: MML_FUNCTION[G, H]) : MML_FUNCTION[G, H] is
			-- The intersection of `current' and `other'.
		require
			other_not_void : other /= Void
			not_continuous : not is_continuous
		deferred
		end

	united (other: MML_FUNCTION[G, H]) : MML_FUNCTION[G, H] is
			-- The union of `current' and `other'.
		require
			other_not_void : other /= Void
			not_continuous : not is_continuous
		deferred
		end

	evaluated (argument : G) : H is
			-- The function `current' evaluated with `argument'.
		require
			argument_not_void : argument /= Void
			argument_defined : is_defined (argument)
		do
			Result := image_element (argument).any_element
		ensure
			definition_of_evaluated : Result.is_equal (image_element (argument).any_element)
		end

feature -- Inversion

	inversed : MML_FUNCTION[H, G] is
			-- The inverse function of `current'.
		require
			must_be_invertible : is_invertible
		deferred
		ensure
			definition_of_inversed : Result.domain.is_equal (range) and Result.range.is_equal (domain)
		end

feature -- Composition

	composed (other : MML_FUNCTION[H, ANY]) : MML_FUNCTION[G, ANY] is
			-- The function composed of `current' and `other'.
		deferred
		end

feature -- Closures

	sequential_closure (other : ANY) : MML_SEQUENCE[ANY] is
			-- The sequenctial closure of `current' starting from `other'.
		require
			other_not_void : other /= Void
			not_continuous : not is_continuous
			same_domain_range : domain.untyped.equals (range)
			same_types : domain.same_type (range)
		deferred
		ensure
			definition_of_sequenctial_closure : untyped.image_element (other).for_all (agent (Result.range).is_member (?))
		end

invariant
	definition_of_function : domain.cardinality.is_equal (cardinality) and is_function_valued and is_relation_valued and is_set_valued

	definition_of_function_inversion : is_invertible implies inversed.inversed.is_equal (Current)

	reflexive_function_union : is_equal (united (Current))
	reflexive_function_intersection : is_equal (intersected (Current))

end -- class MML_FUNCTION