indexing
	description: "Mathematical objects forming graphs over node and edge sets"

deferred class
	MML_GRAPH[G]

inherit
	MML_PAIR[MML_SET[G], MML_RELATION[G, G]]
		rename
			first as nodes,
			second as edges,
			element as component
		export
			{MML_ANY}
				is_identity
		undefine
			is_equal, is_graph_valued
		redefine
			untyped
		end

	MML_TYPEABLE[MML_GRAPH[G]]

feature {NONE} -- Initialization

	make_empty is
			-- Create a new empty graph.
		deferred
		ensure
			definition_of_make_empty : not there_exists (agent nodes.is_member(?)) and not there_exists (agent edges.is_member (?))
		end

feature -- Type

	is_graph_valued : BOOLEAN is
			-- Is `current' graph-valued?
		do
			Result := True
		ensure then
			result_is_true : Result.is_equal (True)
		end

	untyped : MML_GRAPH[ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

feature -- Properties

	is_complete : BOOLEAN is
			-- Is `current' a complete graph?
		do
			Result := nodes.cartesian_product (nodes).is_superset (edges)
		ensure
			definition_of_complete : Result.is_equal (nodes.cartesian_product (nodes).is_superset (edges))
		end

	is_strongly_connected : BOOLEAN is
			-- Is `current' a strongly connected graph?
		do
			Result := edges.reflexive_transitive_closure.is_subset (nodes.cartesian_product (nodes))
		ensure
			definition_of_strongly_connected : Result.is_equal (edges.reflexive_transitive_closure.is_subset (nodes.cartesian_product (nodes)))
		end

	is_weakly_connected : BOOLEAN is
			-- Is `current' a weakly connected graph?
		do
			Result := edges.reflexive_transitive_closure.united (edges.reflexive_transitive_closure.inversed).is_subset (nodes.cartesian_product (nodes))
		ensure
			definition_of_weakly_connected : Result.is_equal (edges.reflexive_transitive_closure.united (edges.reflexive_transitive_closure.inversed).is_subset (nodes.cartesian_product (nodes)))
		end

	is_subgraph (other: MML_GRAPH[G]) : BOOLEAN is
			-- Is `other' a subgraph of `current'?
		require
			other_not_void : other /= Void
		do
			Result := nodes.is_subset (other.nodes) and edges.is_subset (other.edges)
		ensure
			definition_of_subset : Result.is_equal (nodes.is_subset (other.nodes) and edges.is_subset (other.edges))
		end

	is_supergraph (other: MML_GRAPH[G]) : BOOLEAN is
			-- Is `other' a supergraph of `current'?
		require
			other_not_void : other /= Void
		do
			Result := other.is_subgraph (Current)
		ensure
			definition_of_supergraph : Result.is_equal (other.is_subgraph (Current))
		end

	is_proper_supergraph (other: MML_GRAPH[G]) : BOOLEAN is
			-- Is `other' a proper supergraph of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_supergraph (other) and not is_equal (other)
		ensure
			definition_of_proper_supergraph : Result.is_equal (is_supergraph (other) and not is_equal (other))
		end

	is_proper_subgraph (other: MML_GRAPH[G]) : BOOLEAN is
			-- Is `other' a proper subgraph of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_subgraph (other) and not is_equal (other)
		ensure
			definition_of_proper_subgraph : Result.is_equal (is_subgraph (other) and not is_equal (other))
		end

feature -- Basic Operations

	edge_extended (other : MML_PAIR[G, G]) : like Current is
			-- The graph `current' extended with `other'.
		require
			other_not_void : other /= Void
			nodes_in_graph : nodes.is_member (other.first) and nodes.is_member (other.second)
		deferred
		ensure
			definition_of_edge_extended : Result.nodes.is_equal (nodes) and Result.edges.is_equal (edges.extended (other))
		end

	node_extended (other : G) : like Current is
			-- The graph `current' extended with `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_node_extended : Result.nodes.is_equal (nodes.extended (other)) and Result.edges.is_equal (edges)
		end

	edge_pruned (other : MML_PAIR[G, G]) : like Current is
			-- The graph `current' with `other' removed.
		require
			other_not_void : other /= Void
			nodes_in_graph : nodes.is_member (other.first) and nodes.is_member (other.second)
		deferred
		ensure
			definition_of_edge_pruned : Result.nodes.is_equal (nodes) and Result.edges.is_equal (edges.pruned (other))
		end

	node_pruned (other : G) : like Current is
			-- The graph `current' with `other' removed.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_node_pruned : Result.nodes.is_equal (nodes.pruned (other)) and Result.edges.is_equal (edges.domain_element_anti_restricted (other).intersected (edges.range_element_anti_restricted (other)))
		end

invariant
	definition_of_graph : nodes.is_subset (edges.domain) and nodes.is_subset (edges.range)

	strongly_connected_implies_weakly_connected : is_strongly_connected implies is_weakly_connected

end -- class MML_GRAPH