indexing
	description: "Mathematical objects containing two possibly equal elements in a specific order"

deferred class
	MML_PAIR [G, H]

inherit
	MML_ANY
		undefine
			untyped
		redefine
			equals, is_pair_valued
		end

	MML_TYPEABLE[MML_PAIR[G, H]]

	MML_QUANTIFIABLE[ANY]

	MML_PROCESSABLE[ANY]

feature {NONE} -- Initialization

	make_from (one : G; two : H) is
			-- Create a new pair with `one' and `two'.
		require
			one_not_void : one /= Void
			two_not_void : two /= Void
		deferred
		ensure
			first_element_set : first.is_equal (one)
			second_element_set : second.is_equal (two)
		end

	make_from_other (other : MML_PAIR[G, H]) is
			-- Create a new object containing the elements from `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			first_element_set : first.is_equal (other.first)
			second_element_set : second.is_equal (other.second)
		end

feature {MML_ANY} -- Type

	is_compatible (other : MML_ANY) : BOOLEAN is
			-- Is `other' mathematically compatible to `current' ?
		do
			Result := other.conforms_to (Current)
		ensure then
			implementation_of_compatible : other.conforms_to (Current) implies Result
		end

	untyped : MML_PAIR[ANY, ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

feature -- Type

	is_pair_valued : BOOLEAN is
			-- Is `current' pair-valued?
		do
			Result := True
		ensure then
			result_is_true : Result.is_equal (True)
		end

feature -- Comparison

	equals (other: MML_PAIR[G, H]): BOOLEAN is
			-- Is `other' mathematically equivalent to `current' ?
		do
			Result := first.is_equal (other.first) and second.is_equal (other.second)
		ensure then
			implementation_of_equals : Result.is_equal (first.is_equal (other.first) and second.is_equal (other.second))
			equality_implies_compatiblity : Result implies is_compatible (other)
		end

feature -- Access

	element, infix "@" (index : INTEGER) : ANY is
			-- The element of `current' at `index'.
		require
			valid_pair_index : index.is_equal (1) or index.is_equal (2)
		do
			if index = 1 then
				Result := first
			else
				Result := second
			end
		ensure
			definition_of_element : (index.is_equal (1) implies Result.is_equal (first)) and (index.is_equal (2) implies Result.is_equal (second))
		end

	first : G is
			-- The first element of `current'.
		deferred
		ensure
			definition_of_first : element (1).is_equal (Result)
		end

	second : H is
			-- The second element of `current'.
		deferred
		ensure
			definition_of_second : element (2).is_equal (Result)
		end

feature -- Status Report

	is_identity : BOOLEAN is
			-- Does `current' only contain identical elements?
		do
			Result := element (1).is_equal (element (2))
		ensure
			definition_of_identity : Result.is_equal (element (1).is_equal (element (2)))
		end

feature -- Inversion

	inversed : MML_PAIR[H, G] is
			-- The inverse pair of `current'.
		deferred
		ensure
			definition_of_inversed : Result.first.is_equal (second) and Result.second.is_equal (first)
		end

feature -- Processing

	applied (action: PROCEDURE [ANY, TUPLE [ANY]]) : like Current is
			-- The object with `action' applied to every element of `current'.
		do
			Result := clone (Current)
			Result.do_all (action)
		end

feature -- Quantifiers

	there_exists (predicate: FUNCTION [ANY, TUPLE [ANY], BOOLEAN]): BOOLEAN is
			-- Does `current' contain an element which satisfies `predicate'?
		do
			Result := predicate.item ([first])
			if not Result then
				Result := predicate.item ([second])
			end
		ensure then
			definition_of_there_exists : predicate.item ([first]) or predicate.item ([second]) implies Result
		end

	for_all (predicate: FUNCTION [ANY, TUPLE [ANY], BOOLEAN]): BOOLEAN is
			-- Does `current' contain only elements which satisfy `predicate'?
		do
			Result := predicate.item ([first])
			if Result then
				Result := predicate.item ([second])
			end
		ensure then
			definition_of_for_all : predicate.item ([first]) and predicate.item ([second]) implies Result
		end

feature -- Conversion

	identity_pair : MML_PAIR[like Current, like Current] is
			-- The identity pair of `current'.
		deferred
		ensure
			definition_of_identity_pair : Result.first.is_equal (Current) and Result.second.is_equal (Current)
		end

feature {MML_ANY} -- Implementation

	do_all (action: PROCEDURE [ANY, TUPLE [ANY]]) is
			-- Apply `action' to every element of `current'.
		do
			action.call ([first])
			action.call ([second])
		end

	do_if (action: PROCEDURE [ANY, TUPLE [ANY]]; predicate: FUNCTION [ANY, TUPLE [ANY], BOOLEAN]) is
			-- Apply `action' to every item that satisfies `predicate'.
		do
			if predicate.item ([first]) then
				action.call ([first])
			end

			if predicate.item ([second]) then
				action.call ([second])
			end
		end

invariant
	definition_of_pair : is_pair_valued

	definition_of_inversion : is_equal (inversed.inversed )
	definition_of_identity : is_identity implies element (1).is_equal (element (2))

	first_not_void : first /= Void
	second_not_void : second /= Void

end -- class MML_PAIR