indexing
	description: "Mathematical objects that relate elements of a domain set to elements of a range set"

deferred class
	MML_RELATION[G, H]

inherit
	MML_SET[MML_PAIR[G, H]]
		rename
			composed as composed_implementation
		undefine
			is_compatible, is_relation_valued, is_function_valued
		redefine
			united, extended, untyped
		select
			mapped_to_identity
		end

	MML_TYPEABLE[MML_SET[MML_PAIR[G, H]]]

	MML_MAPPABLE[MML_PAIR[G, H], G]
		rename
			mapped as mapped_to_first
		export
			{MML_ANY}
				all
		end

	MML_MAPPABLE[MML_PAIR[G, H], H]
		rename
			mapped as mapped_to_second
		export
			{MML_ANY}
				all
		end

	MML_MAPPABLE[MML_PAIR[G, H], MML_PAIR[H, G]]
		rename
			mapped as mapped_to_inverse
		export
			{MML_ANY}
				all
		end

feature {NONE} -- Initialization

	make_from_sequences (dom: MML_SEQUENCE[G]; ran: MML_SEQUENCE[H]) is
			-- Create a new relation from `dom' and `ran'.
		require
			domain_not_void: dom /= Void
			range_not_void: ran /= Void
			same_count: dom.count.is_equal (ran.count)
		deferred
		ensure
			definition_of_make_from_sequences : domain.is_equal (dom.range) and range.is_equal (ran.range)
			valid_cardinality : cardinality.is_equal (dom.count) and cardinality.is_equal (ran.count)
		end

feature -- Type

	is_relation_valued : BOOLEAN is
			-- Is `current' relation-valued?
		do
			Result := True
		ensure then
			result_is_true : Result.is_equal (True)
		end

	is_function_valued : BOOLEAN is
			-- Is `current' function-valued?
		do
			Result := domain.for_all (agent is_functional (?))
		ensure then
			definition_of_function_valued : domain.for_all (agent is_functional (?))
		end

	untyped : MML_RELATION[ANY, ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

feature -- Status Report

	is_functional (other : G) : BOOLEAN is
			-- Is `current' a functional relation for `other'?
		require
			other_not_void : other /= Void
		do
			Result := domain_element_restricted (other).cardinality <= 1
		ensure
			definition_of_functional : Result.is_equal (domain_element_restricted (other).cardinality <= 1)
		end

feature -- Properties

	is_identity : BOOLEAN is
			-- Is `current' the identity relation?
		do
			Result := domain.untyped.equals (range) and then for_all (agent {MML_PAIR[G, H]}.is_identity)
		ensure
			definition_of_identity : Result.is_equal (domain.untyped.equals (range) and for_all (agent {MML_PAIR[G, H]}.is_identity))
		end

	is_reflexive : BOOLEAN is
			-- Is `current' a reflexive relation?
		do
			Result := domain.untyped.equals (range) and then untyped.is_subset (domain.mapped_to_identity (agent domain.identity_pair_element (?)))
		ensure
			definition_of_reflexitivity : Result.is_equal (domain.untyped.equals (range) and then untyped.is_subset (domain.mapped_to_identity (agent domain.identity_pair_element (?))))
		end

	is_irreflexive : BOOLEAN is
			-- Is `current' an irreflexive relation?
		do
			Result := domain.untyped.equals (range) and then untyped.intersected (domain.mapped_to_identity (agent domain.identity_pair_element (?))).is_empty
		ensure
			definition_of_irreflexitivity : Result.is_equal (domain.untyped.equals (range) and then untyped.intersected (domain.mapped_to_identity (agent domain.identity_pair_element (?))).is_empty)
		end

	is_symmetric : BOOLEAN is
			-- Is `current' a symmetric relation?
		do
			Result := domain.untyped.equals (range) and then inversed.untyped.is_superset (Current)
		ensure
			definition_of_symmetry : Result.is_equal (domain.untyped.equals (range) and  inversed.untyped.is_superset (Current))
		end

	is_asymmetric : BOOLEAN is
			-- Is `current' an asymmetric relation?
		do
			Result := domain.untyped.equals (range) and then inversed.untyped.intersected (Current).is_empty
		ensure
			definition_of_asymmetry : Result.is_equal (domain.untyped.equals (range) and inversed.untyped.intersected (Current).is_empty)
		end

	is_antisymmetric : BOOLEAN is
			-- Is `current' an antisymmetric relation?
		do
			Result := domain.untyped.equals (range) and then inversed.untyped.intersected (Current).is_identity
		ensure
			definition_of_antisymmetry : Result.is_equal (domain.untyped.equals (range) and inversed.untyped.intersected (Current).is_identity)
		end

	is_transitive : BOOLEAN is
			-- Is `current' a transitive relation?
		do
			Result := transitive_closure.is_superset (Current)
		ensure
			definition_of_transitivity : Result.is_equal (transitive_closure.is_superset (Current))
		end

feature -- Basic Operations

	united (other: MML_SET[MML_PAIR[G, H]]) : MML_RELATION[G, H] is
			-- The union of `current' and `other'.
		deferred
		end

	intersected (other: MML_SET[MML_PAIR[G, H]]) : MML_RELATION[G, H] is
			-- The intersection of `current' and `other'.
		deferred
		end

	extended (other: MML_PAIR[G, H]) : MML_RELATION[G, H] is
			-- The relation `current' extended with `other'.
		deferred
		end

feature -- Projections

	domain : MML_SET[G] is
			-- The domain of `current'.
		do
			Result := mapped_to_first (agent {MML_PAIR[G, H]}.first)
		ensure
			definition_of_domain : Result.is_equal (mapped_to_first (agent {MML_PAIR[G, H]}.first))
		end

	range : MML_SET[H] is
			-- The range of `current'.
		do
			Result := mapped_to_second (agent {MML_PAIR[G, H]}.second)
		ensure
			definition_of_range : Result.is_equal (mapped_to_second (agent {MML_PAIR[G, H]}.second))
		end

	image_element (other : G) : MML_SET[H] is
			-- The projected range set of `current' through `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	image (projection : MML_SET[G]) : MML_SET[H] is
			-- The projected range set of `current' through `projection'.
		require
			projection_not_void : projection /= Void
		do
			Result := domain_restricted (projection).range
		ensure
			definition_of_image : Result.is_equal (domain_restricted (projection).range)
		end

	anti_image_element (other : H) : MML_SET[G] is
			-- The projected domain set of `current' through `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	anti_image (projection : MML_SET[H]) : MML_SET[G] is
			-- The projected domain set of `current' through `projection'.
		require
			projection_not_void : projection /= Void
		do
			Result := range_restricted (projection).domain
		ensure
			definition_of_anti_image : Result.is_equal (range_restricted (projection).domain)
		end

	domain_element_restricted (other : G) : like Current is
			-- A relation with pairs of `current' having its first element equivalent to `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	domain_restricted (restriction : MML_SET[G]) : like Current is
			-- A relation with pairs of `current' having its first element in `restriction'.
		require
			restriction_not_void : restriction /= Void
		deferred
		end

	range_element_restricted (other : H) : like Current is
			-- A relation with pairs of `current' having its second element equivalent to `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	range_restricted (restriction : MML_SET[H]) : like Current is
			-- A relation with pairs of `current' having its second element in `restriction'.
		require
			restriction_not_void : restriction /= Void
		deferred
		ensure
			definition_of_range_restricted : composed (restriction.identity_relation).is_equal (Result)
		end

	domain_element_anti_restricted (other : G) : like Current is
			-- A relation with pairs of `current' having its first element not equivalent to `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	domain_anti_restricted (restriction : MML_SET[G]) : like Current is
			-- A relation with pairs of `current' having its first element not in `restriction'.
		require
			restriction_not_void : restriction /= Void
		do
			Result := domain_restricted (domain.subtracted (restriction))
		ensure
			definition_of_domain_anti_restricted : Result.is_equal (domain_restricted (domain.subtracted (restriction)))
		end

	range_element_anti_restricted (other : H) : like Current is
			-- A relation with pairs of `current' having its second element not equivalent `other'.
		require
			other_not_void : other /= Void
		deferred
		end

	range_anti_restricted (restriction : MML_SET[H]) : like Current is
			-- A relation with pairs of `current' having its second element not in `restriction'.
		require
			restriction_not_void : restriction /= Void
		do
			Result := range_restricted (range.subtracted (restriction))
		ensure
			definition_of_range_anti_restricted : Result.is_equal (range_restricted (range.subtracted (restriction)))
		end

feature -- Inversion

	inversed : MML_RELATION[H, G] is
			-- The inverse relation of `current'.
		deferred
		ensure
			definition_of_inversed : Result.domain.is_equal (range) and Result.range.is_equal (domain)
			same_cardinality : Result.cardinality.is_equal (cardinality)
		end

feature -- Composition

	composed (other : MML_RELATION[H, ANY]) : MML_RELATION[G, ANY] is
			-- The relation composed of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_composed : Result.domain.is_superset (anti_image (range.intersected (other.domain))) and Result.range.is_superset (other.image (range.intersected (other.domain)))
			same_domain_range : Result.range.same_type (other.range)
		end

feature -- Closures

	transitive_closure : MML_RELATION[ANY, ANY] is
			-- The transitive closure of `current'.
		require
			same_types : domain.same_type (range)
		deferred
		ensure
			definition_of_transitive_closure : untyped.is_superset (Result)
		end

	reflexive_transitive_closure : MML_RELATION[ANY, ANY] is
			-- The reflexive-transitive closure of `current'.
		require
			same_types : domain.same_type (range)
		deferred
		ensure
			definition_of_reflexive_transitive_closure : Result.is_subset (transitive_closure) and untyped.is_superset (Result)
		end

invariant
	definition_of_relation : for_all (agent {MML_PAIR[G, H]}.is_pair_valued) and is_set_valued and is_relation_valued

	definition_of_relation_inversion : inversed.inversed.is_equal (Current)
	definition_of_domain_range : inversed.domain.equals (range) and inversed.range.equals (domain)
	definition_of_composition : inversed.composed (Current).is_identity

	identity_implies_reflexitivity : is_identity implies is_reflexive

end -- class MML_RELATION