indexing
	description: "Mathematical objects that contain unique elements of the same type in unspecified order"

deferred class
	MML_SET[G]

inherit
	MML_ANY
		redefine
			equals, untyped, is_set_valued
		end

	MML_TYPEABLE[MML_SET[G]]

	MML_QUANTIFIABLE[G]

	MML_PROCESSABLE[G]

	MML_FILTERABLE[G]

	MML_CONVERTIBLE[G]
		rename
			identity_pair as identity_pair_element
		end

	MML_MAPPABLE[G, MML_PAIR[G, G]]
		rename
			mapped as mapped_to_identity
		export
			{MML_ANY}
				all
		end

	MML_EXTENDIBLE[G]
		export
			{MML_ANY}
				all
		end

feature {NONE} -- Initialization

	make_from_element (other : G) is
			-- Create a new set containing the element `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_make_from : is_member (other)
			valid_cardinality : cardinality.is_equal (1)
		end

	make_from_other (other : MML_SET[G]) is
			-- Create a new object containing the elements from `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_make_from : is_equal (other)
			same_members : other.for_all (agent is_member(?))
			same_cardinality : cardinality.is_equal (other.cardinality)
		end

	make_from_set (other : MML_SET[G]) is
			-- Create a new set containing the elements from `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_make_from : is_equal (other)
			same_members : other.for_all (agent is_member(?))
			same_cardinality : cardinality.is_equal (other.cardinality)
		end

	make_empty is
			-- Create a new empty set.
		deferred
		ensure
			definition_of_make_empty : not there_exists (agent is_member(?))
			valid_cardinality : cardinality.is_equal (0)
		end

feature {MML_ANY} -- Type

	is_compatible (other : MML_ANY) : BOOLEAN is
			-- Is `other' mathematically compatible to `current' ?
		do
			Result := other.conforms_to (Current)
		ensure then
			implementation_of_compatible : other.conforms_to (Current) implies Result
		end

feature -- Type

	untyped : MML_SET[ANY] is
			-- The object `current' as typeless object.
		do
			Result := Current
		end

	is_set_valued : BOOLEAN is
			-- Is `current' set-valued?
		do
			Result := True
		ensure then
			result_is_true : Result.is_equal (True)
		end

feature -- Comparison

	equals (other: MML_SET[G]): BOOLEAN is
			-- Is `other' mathematically equivalent to `current'?
		do
			Result := (cardinality = other.cardinality) and then other.for_all (agent is_member (?))
		ensure then
			implementation_of_equals : Result.is_equal ((cardinality.is_equal (other.cardinality)) and other.for_all (agent is_member (?)))
		end

feature -- Access

	any_element : G is
			-- An arbitrary element of `current'.
		require
			not_empty : not is_empty
		deferred
		ensure
			definition_of_any_element : is_member (Result)
		end

	any_item (predicate: FUNCTION [ANY, TUPLE [G], BOOLEAN]) : G is
			-- An arbitrary element of `current' which satisfies `predicate'?
		require
			predicate_not_void : predicate /= Void
			there_exists_item : there_exists (predicate)
		deferred
		ensure
			definition_of_any_item : is_member (Result) and predicate.item ([Result])
		end

feature -- Conversion

	identity_relation : MML_RELATION[G, G] is
			-- The identity relation of `current'.
		deferred
		ensure
			definition_of_identity_relation : Result.for_all (agent {MML_PAIR[G, G]}.is_identity)
			same_domain_range : is_equal (Result.domain) and is_equal (Result.range)
			same_cardinality : Result.cardinality.is_equal (cardinality)
		end

	power_set : MML_POWERSET[G] is
			-- The set `current' as power set.
		deferred
		ensure
			definition_of_powerset : Result.is_member (Current)
			valid_cardinality : Result.cardinality.is_equal (1)
		end

	random_sequence : MML_SEQUENCE[G] is
			-- The set `current' as random sequence.
		deferred
		ensure
			definition_of_random_sequence : Result.for_all (agent is_member (?)) and is_equal (Result.range)
			same_cardinality : Result.count.is_equal (cardinality)
		end

feature -- Status Report

	cardinality : INTEGER is
			-- The cardinality of `current'
		deferred
		ensure
			non_negative_cardinality : Result >= 0
		end

feature -- Properties

	is_member (other: G) : BOOLEAN is
			-- Is `other' a member of `current'?
		require
			other_not_void : other /= Void
		deferred
		ensure
			empty_implies_not_member : is_empty implies not Result
		end

	is_empty : BOOLEAN is
			-- Is `current' empty?
		do
			Result := (cardinality = 0)
		ensure
			definition_of_empty : Result.is_equal (cardinality.is_equal (0))
		end

	is_disjoint (other: MML_SET[G]) : BOOLEAN is
			-- Is `other' disjoint from `current'?
		require
			other_not_void: other /= Void
		do
			Result := intersected (other).is_empty
		ensure
			definition_of_disjoint: Result.is_equal (intersected (other).is_empty)
			empty_implies_disjoint: (is_empty or other.is_empty) implies Result
		end

	is_subset (other: MML_SET[G]) : BOOLEAN is
			-- Is `other' a subset of `current'?
		require
			other_not_void : other /= Void
		do
			Result := other.for_all (agent is_member(?))
		ensure
			definition_of_subset : Result.is_equal (other.for_all (agent is_member (?)))
			valid_cardinality : Result.is_equal (cardinality >= other.cardinality)
		end

	is_superset (other: MML_SET[G]) : BOOLEAN is
			-- Is `other' a superset of `current'?
		require
			other_not_void : other /= Void
		do
			Result := other.is_subset (Current)
		ensure
			definition_of_superset : Result.is_equal (other.is_subset (Current))
			valid_cardinality : Result.is_equal (other.cardinality >= cardinality)
		end

	is_proper_superset (other: MML_SET[G]) : BOOLEAN is
			-- Is `other' a proper superset of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_superset (other) and not is_equal (other)
		ensure
			definition_of_proper_superset : Result.is_equal (is_superset (other) and not is_equal (other))
			all_elements_included : Result implies for_all (agent other.is_member(?))
			there_exists_element : Result implies other.there_exists (agent negated (?, agent is_member (?)))
			valid_cardinality : Result.is_equal (other.cardinality > cardinality)
		end

	is_proper_subset (other: MML_SET[G]) : BOOLEAN is
			-- Is `other' a proper subset of `current'?
		require
			other_not_void : other /= Void
		do
			Result := is_subset (other) and not is_equal (other)
		ensure
			definition_of_proper_subset : Result.is_equal (is_subset (other) and not is_equal (other))
			all_elements_included : Result implies other.for_all (agent is_member(?))
			there_exists_element : Result implies there_exists (agent negated (?, agent other.is_member(?)))
			valid_cardinality : Result.is_equal (other.cardinality < cardinality)
		end

feature -- Basic Operations

	cartesian_product (other : MML_SET[ANY]) : MML_SET[MML_PAIR[G, ANY]] is
			-- The cartesian product of `other' and `current'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_cartesian_product : Result.cardinality.is_equal (cardinality * other.cardinality)
		end

	intersected (other: MML_SET[G]) : like Current is
			-- The intersection of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_intersected : Result.for_all (agent anded (?, agent is_member (?), agent other.is_member (?)))
			generalized_intersected : power_set.extended (other).generalized_intersected.is_equal (Result)
			existing_elements : filtered (agent other.is_member (?)).is_equal (Result)
			other_elements : other.filtered (agent is_member (?)).is_equal (Result)
			existing_subset : is_subset (Result)
			other_subset : other.is_subset (Result)
			valid_cardinality : Result.cardinality <= cardinality and Result.cardinality <= other.cardinality
		end

	united (other: MML_SET[G]) : MML_SET[G] is
			-- The union of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_united : power_set.extended (other).for_all_sets (agent Result.is_subset (?))
			generalized_united : power_set.extended (other).generalized_united.is_equal (Result)
			existing_elements : for_all (agent Result.is_member (?))
			other_elements : other.for_all (agent Result.is_member (?))
			valid_cardinality : Result.cardinality >= cardinality and Result.cardinality >= other.cardinality
		end

	subtracted (other: MML_SET[G]) : like Current is
			-- The difference of `current' and `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_subtracted : Result.is_equal (filtered (agent negated (?, agent other.is_member (?))))
			disjoint_subtracted_result : Result.intersected (other).is_empty
			valid_cardinality : Result.cardinality <= cardinality
		end

	difference (other: MML_SET[G]) : like Current is
			-- The symmetric difference of `current' and `other'.
		require
			other_not_void : other /= Void
		do
			Result := united (other).subtracted (intersected (other))
		ensure
			definition_of_difference : Result.is_equal (united (other).subtracted (intersected (other)))
		end

	extended (other: G) : like Current is
			-- The set `current' extended with `other'.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_extended : Result.is_member (other)
			other_members_remain : for_all (agent Result.is_member (?))
			has_implies_same : is_member (other) implies is_equal (old Current)
			valid_cardinality : Result.cardinality >= cardinality
		end

	pruned (other: G) : like Current is
			-- The set `current' with `other' removed.
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_pruned : is_equal (Result.extended (other))
			no_member_anymore : not Result.is_member (other)
			not_has_implies_same : not is_member (other) implies is_equal (old Current)
			valid_cardinality : Result.cardinality <= cardinality
		end

feature -- Partition

	is_partition (other : MML_SET[MML_SET[G]]) : BOOLEAN is
			-- Is `other' a partition of `current'?
		require
			other_not_void : other /= Void
		deferred
		ensure
			definition_of_partition : Result implies other.for_all (agent is_subset (?))
		end

invariant
	definition_of_set : is_set_valued
	definition_of_empty_set : is_empty.is_equal (cardinality.is_equal (0))

	reflexive_subset_relation : is_subset (Current)
	reflexive_superset_relation: is_superset (Current)

	definition_of_union : is_equal (united (Current))
	definition_of_intersection : is_equal (intersected (Current))

	non_negative_cardinality : cardinality >= 0

end -- class MML_SET