deferred class SPECIFICATION_DISPENSER [G]

inherit
	SPECIFICATION_ACTIVE [G]

	SPECIFICATION_FINITE [G]

	MML_SPECIFICATION

feature {MML_SPECIFICATION, ANY} -- Model

	sequence_model : MML_SEQUENCE[G] is
			-- The sequence model
		deferred
		ensure
			definition_of_sequence_model: Result.range.for_all (agent has (?))
			pure_feature: is_equal (old Current)
			result_not_void: Result /= Void
		end

	cursor_model : INTEGER is
			-- The cursor model
		deferred
		ensure
			definition_of_cursor_model: Result >= sequence_model.lower_bound - 1 and Result <= sequence_model.upper_bound + 1
			pure_feature: is_equal (old Current)
			result_not_void: Result /= Void
		end

feature -- Access

	has (other: G): BOOLEAN is
			-- Does structure include `other'?
		deferred
		ensure then
			definition_of_has: Result.is_equal (sequence_model.is_member (other))
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	item: G is
			-- Current item
		deferred
		ensure then
			definition_of_item: Result.is_equal (sequence_model.element (cursor_model))
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

feature -- Measurement

	count: INTEGER is
			-- Number of items
		deferred
		ensure then
			definition_of_count: Result.is_equal (sequence_model.count) and Result.is_equal (bag_model.cardinality)
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model) and bag_model.is_equal (old bag_model)
		end

	occurrences (other: G): INTEGER is
			-- Number of times `other' appears in structure
		deferred
		ensure then
			definition_of_occurrences: Result.is_equal (sequence_model.indexes_element (other).cardinality)
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		deferred
		ensure then
			definition_of_is_equal: Result.is_equal (sequence_model.is_equal (other.sequence_model))
			immutable_model: sequence_model.is_equal (old sequence_model)
		end

feature -- Status report

	changeable_comparison_criterion: BOOLEAN is
			-- May `object_comparison' be changed?
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	full: BOOLEAN is
			-- Is structure filled to capacity?
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model) and bag_model.is_equal (old bag_model)
		end

	is_empty: BOOLEAN is
			-- Is structure empty?
		deferred
		ensure then
			definition_of_is_empty: Result.is_equal (sequence_model.is_empty)
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	object_comparison: BOOLEAN is
			-- Must search operations use `equal' rather than `='
			-- for comparing references? (Default: no, use `='.)
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

feature -- Element change

	append (other: SPECIFICATION_SEQUENCE [G]) is
			-- Append a copy of `other'.
		require
			other_not_void: other /= Void
		deferred
		ensure
			definition_of_append: old extendible implies sequence_model.is_equal (old sequence_model.concatenated (other.sequence_model))
			definition_of_append: old extendible implies bag_model.is_subbag (old bag_model.united (other.bag_model))
			immutable_argument: other.is_equal (old other)
			extendible_implies_filled_in: (old extendible) and not other.linear_representation.is_empty implies not is_empty
			not_extendible_implies_untouched: not (old extendible) implies is_equal (old Current)
		end

	replace (other: G) is
			-- Replace current item by `other'.
		deferred
		ensure then
			definition_of_replace: sequence_model.is_equal (old sequence_model.replaced (other, cursor_model))
			immutable_model: cursor_model.is_equal (old cursor_model)
		end

feature -- Removal

	prune (other: G) is
			-- Remove one occurrence of `other' if any.
		deferred
		ensure then
			definition_of_prune: sequence_model.is_equal (old sequence_model.pruned_at (sequence_model.indexes_element_ordered (other).first))
			immutable_model: cursor_model.is_equal (old cursor_model)
		end

	prune_all (other: G) is
			-- Remove all occurrences of `other'.
		deferred
		ensure then
			definition_of_prune: sequence_model.is_equal (old sequence_model.pruned_at (sequence_model.indexes_element_ordered (other).first))
			immutable_model: cursor_model.is_equal (old cursor_model)
		end

	remove is
			-- Remove current item.
		deferred
		ensure then
			definition_of_remove: sequence_model.is_equal (old sequence_model.pruned_at (cursor_model))
		end

	wipe_out is
			-- Remove all items.
		deferred
		ensure then
			definition_of_wipe_out: sequence_model.is_empty
			undefined_cursor: not sequence_model.is_defined (cursor_model)
		end

feature -- Status Report

	extendible: BOOLEAN is
			-- May new items be added?)
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	is_inserted (other: G): BOOLEAN is
			-- Has `other' been inserted by the most recent insertion?
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	prunable: BOOLEAN is
			-- May items be removed?
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

feature -- Status Setting

	compare_objects is
			-- Ensure that future search operations will use `equal'
			-- rather than `=' for comparing references.
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	compare_references is
			-- Ensure that future search operations will use `='
			-- rather than `equal' for comparing references.
		deferred
		ensure then
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

feature -- Status report

	readable: BOOLEAN is
			-- Is there a current item that may be read?
		deferred
		ensure then
			definition_of_readable: Result.is_equal (not sequence_model.is_empty)
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

	writable: BOOLEAN is
			-- Is there a current item that may be modified?
		deferred
		ensure then
			definition_of_writable: Result.is_equal (not sequence_model.is_empty)
			immutable_models: sequence_model.is_equal (old sequence_model) and cursor_model.is_equal (old cursor_model)
		end

invariant
	definition_of_readable: readable.is_equal (not sequence_model.is_empty)
	definition_of_writable: writable.is_equal (not sequence_model.is_empty)

	writable_implies_readable: writable implies readable

	glueing_invariant: bag_model.domain.is_equal (sequence_model.range)
	glueing_invariant: cursor_model >= sequence_model.lower_bound - 1 and cursor_model <= sequence_model.upper_bound + 1

end -- class SPECIFICATION_DISPENSER