deferred class 
	SPECIFICATION_LINEAR_SUBSET [G]

inherit
	SPECIFICATION_TRAVERSABLE_SUBSET[G]
		rename
			cursor_model as cursor,
			sequence_model as sequence,
			set_model as set
		end

feature -- Access

	index: INTEGER is
			-- Current index
		deferred
		ensure
			Result >= sequence.lower_bound
			Result <= sequence.upper_bound + 1
			Result.is_equal (cursor)
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered equal to current object?
		deferred
		ensure then
			Result implies sequence.is_equal (other.sequence)
			Result implies set.is_equal (other.set)
		end

	is_subset (other: SPECIFICATION_SUBSET [G]): BOOLEAN is
			-- Is current set a subset of `other'?
		deferred
		ensure then
			sequence.is_equal (old sequence) and cursor.is_equal (old cursor)
		end

	is_superset (other: SPECIFICATION_SUBSET [G]): BOOLEAN is
			-- Is current set a superset of `other'?
		deferred
		ensure then
			sequence.is_equal (old sequence) and cursor.is_equal (old cursor)
		end

feature -- Status report

	before: BOOLEAN is
			-- Is cursor at left from first item?
		deferred
		ensure
			Result.is_equal (not sequence.is_defined (cursor))
			cursor < sequence.lower_bound
			is_equal (old Current)
			sequence.equals (old sequence)
			cursor.is_equal (old cursor)
			set.equals (old set)
		end

	islast: BOOLEAN is
			-- Is cursor at last item?
		deferred
		ensure
			Result.is_equal (cursor.is_equal (sequence.upper_bound))
			is_equal (old Current)
			sequence.equals (old sequence)
			cursor.is_equal (old cursor)
			set.equals (old set)
			Result implies sequence.is_defined (cursor)
		end

	off: BOOLEAN is
			-- Is there no current item?
		deferred
		ensure then
			Result.is_equal (not sequence.is_defined (cursor))
		end

	valid_index (number: INTEGER): BOOLEAN is
			-- Is `number' a valid index?
		deferred
		ensure
			Result.is_equal (number >= sequence.lower_bound - 1)
			Result.is_equal (number <= sequence.upper_bound + 1)
			is_equal (old Current)
			number.is_equal (old number)
			sequence.equals (old sequence)
			cursor.is_equal (old cursor)
			set.equals (old set)
		end

feature -- Cursor movement

	go_i_th (i: INTEGER) is
			-- Move cursor to `i'-th item.
		require
			valid_index (i)
			i >= sequence.lower_bound - 1 and i <= sequence.upper_bound + 1
		deferred
		ensure
			cursor.is_equal (i)
			i.is_equal (old i)
			sequence.equals (old sequence) and set.equals (old set)
			i < sequence.lower_bound implies before
			i > sequence.upper_bound implies after
			cursor.is_equal (i)
		end

feature -- Element change

	extend (other: G) is
			-- Ensure that set includes `other'.
		deferred
		ensure then
			sequence.equals ((old sequence).extended_at (other, (old sequence).upper_bound))
			cursor.is_equal (old cursor)
		end

	merge (other: SPECIFICATION_SUBSET [G]) is
			-- Add all items of `other'.
		deferred
		ensure then
			sequence.interval (sequence.lower_bound, (old sequence).upper_bound).equals (old sequence)
			cursor.is_equal (old cursor)
		end

	put_left (other: G) is
			-- Insert `other' before the cursor.
		require else
			other /= void
			not before
		deferred
		ensure then
			index = old index + 1
		end
	
feature -- Removal

	prune_all (other: G) is
			-- Remove all occurrences of `other'.
		deferred
		ensure then
			set.equals ((old set).pruned (other))
		end

	wipe_out is
			-- Remove all items.
		deferred
		ensure then
			sequence.is_empty
			set.is_empty
			cursor > sequence.upper_bound
			not sequence.is_defined (cursor)
		end

feature -- Conversion

	linear_representation: SPECIFICATION_LINEAR [G] is
			-- Representation as a linear structure
		deferred
		ensure then
			Result.for_all (agent set.is_member (?))
			Result.for_all (agent sequence.is_member (?))
			sequence.equals (old sequence)
			cursor.is_equal (old cursor)
			set.equals (old set)
		end

feature -- Duplication

	duplicate (number: INTEGER): like Current is
			-- New structure containing min (`number', `count') items from current structure
		deferred
		ensure then
			Result.sequence.for_all (agent sequence.is_member (?))
			number <= count implies Result.sequence.count.is_equal (number)
			number >= count implies Result.sequence.count.is_equal (count)
			sequence.equals (old sequence)
			cursor.is_equal (old cursor)
		end

feature -- Element Change

	fill (other: SPECIFICATION_CONTAINER [G]) is
			-- Fill with as many items of `other' as possible.
		deferred
		ensure then
			sequence.interval (sequence.lower_bound, (old sequence).upper_bound).equals (old sequence)
			other.linear_representation.for_all (agent set.is_member (?))
			cursor.is_equal (old cursor)
		end

feature -- Status Report

	is_inserted (other: G): BOOLEAN is
			-- Has `other' been inserted by the most recent insertion?
		deferred
		ensure then
			Result implies set.is_member (other)
			Result implies sequence.is_member (other)
		end

invariant
	before.is_equal (cursor.is_equal (0))

end -- class SPECIFICATION_LINEAR_SUBSET